@@ -558,16 +558,16 @@ rip_rte_process (struct rte *rte, struct sockaddr_in *from,
 	      rinfo->flags |= RIP_RTF_FIB;
 	    }
 
-
-	  if (! IPV4_ADDR_SAME (&rinfo->nexthop, nexthop) &&
-	      oldmetric != RIP_METRIC_INFINITY)
+	  /* Update nexthop and/or metric value.  */
+	  if (oldmetric != RIP_METRIC_INFINITY)
 	    {
 	      rip_zebra_ipv4_delete (&p, &rinfo->nexthop, oldmetric);
 	      rip_zebra_ipv4_add (&p, nexthop, rinfo->metric,
 				  rinfo->distance);
 	      rinfo->flags |= RIP_RTF_FIB;
 
-	      IPV4_ADDR_COPY (&rinfo->nexthop, nexthop);
+	      if (! IPV4_ADDR_SAME (&rinfo->nexthop, nexthop))
+		IPV4_ADDR_COPY (&rinfo->nexthop, nexthop);
 	    }
 
 	  /* - Set the route change flag and signal the output process
@@ -627,7 +627,6 @@ rip_packet_dump (struct rip_packet *packet, int size, char *sndrcv)
   u_char netmask = 0;
   u_char *p;
 
-
   /* Set command string. */
   if (packet->command > 0 && packet->command < RIP_COMMAND_MAX)
     command_str = lookup (rip_msg, packet->command);
@@ -667,7 +666,7 @@ rip_packet_dump (struct rip_packet *packet, int size, char *sndrcv)
 		  zlog_info ("    RIP-2 packet len %d Key ID %d"
 			     " Auth Data len %d", ntohs (md5->packet_len),
 			     md5->keyid, md5->auth_len);
-		  zlog_info ("    Sequence Number %d", ntohl (md5->sequence));
+		  zlog_info ("    Sequence Number %ld", (u_long)ntohl (md5->sequence));
 		}
 	      else if (ntohs (rte->tag) == RIP_AUTH_DATA)
 		{
@@ -687,18 +686,18 @@ rip_packet_dump (struct rip_packet *packet, int size, char *sndrcv)
 		}
             }
 	  else
-	    zlog_info ("  %s/%d -> %s family %d tag %d metric %d",
+	    zlog_info ("  %s/%d -> %s family %d tag %d metric %ld",
 		       inet_ntop (AF_INET, &rte->prefix, pbuf, BUFSIZ),netmask,
 		       inet_ntop (AF_INET, &rte->nexthop, nbuf, BUFSIZ),
 		       ntohs (rte->family), ntohs (rte->tag), 
-		       ntohl (rte->metric));
+		       (u_long)ntohl (rte->metric));
 	}
       else
 	{
-	  zlog_info ("  %s family %d tag %d metric %d", 
+	  zlog_info ("  %s family %d tag %d metric %ld", 
 		     inet_ntop (AF_INET, &rte->prefix, pbuf, BUFSIZ),
 		     ntohs (rte->family), ntohs (rte->tag),
-		     ntohl (rte->metric));
+		     (u_long)ntohl (rte->metric));
 	}
     }
 }
@@ -1231,6 +1230,14 @@ rip_redistribute_add (int type, int sub_type, struct prefix_ipv4 *p,
 
   if (rinfo)
     {
+      if (rinfo->type == ZEBRA_ROUTE_CONNECT 
+	  && rinfo->sub_type == RIP_ROUTE_INTERFACE
+	  && rinfo->metric != RIP_METRIC_INFINITY)
+	{
+	  route_unlock_node (rp);
+	  return;
+	}
+
       /* Manually configured RIP route check. */
       if (rinfo->type == ZEBRA_ROUTE_RIP 
 	  && rinfo->sub_type == RIP_ROUTE_STATIC)
@@ -1287,7 +1294,6 @@ rip_redistribute_delete (int type, int sub_type, struct prefix_ipv4 *p,
     return;
 
   rp = route_node_lookup (rip->table, (struct prefix *) p);
-
   if (rp)
     {
       rinfo = rp->info;
@@ -1368,7 +1374,6 @@ rip_request_process (struct rip_packet *packet, int size,
 	  apply_mask_ipv4 (&p);
 	  
 	  rp = route_node_lookup (rip->table, (struct prefix *) &p);
-
 	  if (rp)
 	    {
 	      rinfo = rp->info;
@@ -1557,7 +1562,7 @@ rip_read (struct thread *t)
     }
 
   /* Dump RIP packet. */
-  if (IS_RIP_DEBUG_PACKET)
+  if (IS_RIP_DEBUG_RECV)
     rip_packet_dump (packet, len, "RECV");
 
   /* RIP version adjust.  This code should rethink now.  RFC1058 says
@@ -1752,7 +1757,7 @@ rip_create_socket ()
   struct sockaddr_in addr;
   struct servent *sp;
 
-  bzero (&addr, sizeof (struct sockaddr_in));
+  memset (&addr, 0, sizeof (struct sockaddr_in));
 
   /* Set RIP port. */
   sp = getservbyname ("router", "udp");
@@ -1994,9 +1999,24 @@ rip_output_process (struct interface *ifp, struct sockaddr_in *to,
 	rinfo->nexthop_out.s_addr = 0;
 	rinfo->metric_out = rinfo->metric;
 	rinfo->ifindex_out = ifp->ifindex;
+
+	/* In order to avoid some local loops, if the RIP route has a
+	   nexthop via this interface, keep the nexthop, otherwise set
+	   it to 0. The nexthop should not be propagated beyond the
+	   local broadcast/multicast area in order to avoid an IGP
+	   multi-level recursive look-up.  For RIP and connected
+	   route, we don't set next hop value automatically.  For
+	   settting next hop to those routes, please use
+	   route-map.  */
+
+	if (rinfo->type != ZEBRA_ROUTE_RIP
+	    && rinfo->type != ZEBRA_ROUTE_CONNECT
+	    && rinfo->ifindex == ifp->ifindex)
+	  rinfo->nexthop_out = rinfo->nexthop;
            
 	/* Apply route map - continue, if deny */
-	if (rip->route_map[rinfo->type].name) 
+	if (rip->route_map[rinfo->type].name
+	    && rinfo->sub_type != RIP_ROUTE_INTERFACE)
 	  {
 	    ret = route_map_apply (rip->route_map[rinfo->type].map,
 				   (struct prefix *)p, RMAP_RIP, rinfo);
@@ -2047,7 +2067,7 @@ rip_output_process (struct interface *ifp, struct sockaddr_in *to,
 	    ret = rip_send_packet (STREAM_DATA (s), stream_get_endp (s),
 				   to, ifp);
 
-	    if (ret >= 0 && IS_RIP_DEBUG_PACKET)
+	    if (ret >= 0 && IS_RIP_DEBUG_SEND)
 	      rip_packet_dump ((struct rip_packet *)STREAM_DATA (s),
 			       stream_get_endp(s), "SEND");
 	    num = 0;
@@ -2063,7 +2083,7 @@ rip_output_process (struct interface *ifp, struct sockaddr_in *to,
 
       ret = rip_send_packet (STREAM_DATA (s), stream_get_endp (s), to, ifp);
 
-      if (ret >= 0 && IS_RIP_DEBUG_PACKET)
+      if (ret >= 0 && IS_RIP_DEBUG_SEND)
 	rip_packet_dump ((struct rip_packet *)STREAM_DATA (s),
 			 stream_get_endp (s), "SEND");
       num = 0;
@@ -2319,7 +2339,8 @@ rip_redistribute_withdraw (int type)
   for (rp = route_top (rip->table); rp; rp = route_next (rp))
     if ((rinfo = rp->info) != NULL)
       {
-	if (rinfo->type == type)
+	if (rinfo->type == type
+	    && rinfo->sub_type != RIP_ROUTE_INTERFACE)
 	  {
 	    /* Perform poisoned reverse. */
 	    rinfo->metric = RIP_METRIC_INFINITY;
@@ -2386,7 +2407,7 @@ rip_request_send (struct sockaddr_in *to, struct interface *ifp,
 }
 
 int
-rip_update_jitter (int time)
+rip_update_jitter (unsigned long time)
 {
   return ((rand () % (time + 1)) - (time / 2));
 }
@@ -2555,7 +2576,6 @@ DEFUN (no_rip_route,
 
   /* For router rip configuration. */
   node = route_node_lookup (rip->route, (struct prefix *) &p);
-
   if (! node)
     {
       vty_out (vty, "Can't find route %s.%s", argv[0],
@@ -2622,7 +2642,7 @@ ALIAS (no_rip_default_metric,
 
 DEFUN (rip_timers,
        rip_timers_cmd,
-       "timers basic <0-4294967295> <1-4294967295> <1-4294967295>",
+       "timers basic <5-2147483647> <5-2147483647> <5-2147483647>",
        "Adjust routing timers\n"
        "Basic routing protocol update timers\n"
        "Routing table update timer value in second. Default is 30.\n"
@@ -2633,23 +2653,25 @@ DEFUN (rip_timers,
   unsigned long timeout;
   unsigned long garbage;
   char *endptr = NULL;
+  unsigned long RIP_TIMER_MAX = 2147483647;
+  unsigned long RIP_TIMER_MIN = 5;
 
   update = strtoul (argv[0], &endptr, 10);
-  if (update == ULONG_MAX || *endptr != '\0')
+  if (update > RIP_TIMER_MAX || update < RIP_TIMER_MIN || *endptr != '\0')  
     {
       vty_out (vty, "update timer value error%s", VTY_NEWLINE);
       return CMD_WARNING;
     }
   
   timeout = strtoul (argv[1], &endptr, 10);
-  if (timeout == ULONG_MAX || *endptr != '\0')
+  if (timeout > RIP_TIMER_MAX || timeout < RIP_TIMER_MIN || *endptr != '\0') 
     {
       vty_out (vty, "timeout timer value error%s", VTY_NEWLINE);
       return CMD_WARNING;
     }
   
   garbage = strtoul (argv[2], &endptr, 10);
-  if (garbage == ULONG_MAX || *endptr != '\0')
+  if (garbage > RIP_TIMER_MAX || garbage < RIP_TIMER_MIN || *endptr != '\0') 
     {
       vty_out (vty, "garbage timer value error%s", VTY_NEWLINE);
       return CMD_WARNING;
@@ -2984,6 +3006,26 @@ rip_vty_out_uptime (struct vty *vty, struct rip_info *rinfo)
     }
 }
 
+char *
+rip_route_type_print (int sub_type)
+{
+  switch (sub_type)
+    {
+      case RIP_ROUTE_RTE:
+	return "n";
+      case RIP_ROUTE_STATIC:
+	return "s";
+      case RIP_ROUTE_DEFAULT:
+	return "d";
+      case RIP_ROUTE_REDISTRIBUTE:
+	return "r";
+      case RIP_ROUTE_INTERFACE:
+	return "i";
+      default:
+	return "?";
+    }
+}
+
 DEFUN (show_ip_rip,
        show_ip_rip_cmd,
        "show ip rip",
@@ -2997,24 +3039,24 @@ DEFUN (show_ip_rip,
   if (! rip)
     return CMD_SUCCESS;
 
-  vty_out (vty, "Codes: R - RIP, C - connected, O - OSPF, B - BGP%s%s"
-	   "   Network            Next Hop         Metric From            Time%s",
-	   VTY_NEWLINE,
-	   VTY_NEWLINE,
-	   VTY_NEWLINE);
+  vty_out (vty, "Codes: R - RIP, C - connected, O - OSPF, B - BGP%s"
+	   "      (n) - normal, (s) - static, (d) - default, (r) - redistribute,%s"
+	   "      (i) - interface%s%s"
+	   "     Network            Next Hop         Metric From            Time%s",
+	   VTY_NEWLINE, VTY_NEWLINE, VTY_NEWLINE, VTY_NEWLINE, VTY_NEWLINE);
   
   for (np = route_top (rip->table); np; np = route_next (np))
     if ((rinfo = np->info) != NULL)
       {
 	int len;
 
-	len = vty_out (vty, "%s%s %s/%d",
+	len = vty_out (vty, "%s(%s) %s/%d",
 		       /* np->lock, For debugging. */
 		       route_info[rinfo->type].str,
-		       " ",
+		       rip_route_type_print (rinfo->sub_type),
 		       inet_ntoa (np->p.u.prefix4), np->p.prefixlen);
 	
-	len = 22 - len;
+	len = 24 - len;
 
 	if (len > 0)
 	  vty_out (vty, "%*s", len, " ");
@@ -3023,7 +3065,7 @@ DEFUN (show_ip_rip,
 	  vty_out (vty, "%-20s %2d ", inet_ntoa (rinfo->nexthop),
 		   rinfo->metric);
         else
-	  vty_out (vty, "                     %2d ", rinfo->metric);
+	  vty_out (vty, "0.0.0.0              %2d ", rinfo->metric);
 
 	/* Route which exist in kernel routing table. */
 	if ((rinfo->type == ZEBRA_ROUTE_RIP) && 
@@ -3034,9 +3076,11 @@ DEFUN (show_ip_rip,
 	  }
 	else if (rinfo->metric == RIP_METRIC_INFINITY)
 	  {
-	    vty_out (vty, "%-15s ", "");
+	    vty_out (vty, "self            ");
 	    rip_vty_out_uptime (vty, rinfo);
 	  }
+	else
+	  vty_out (vty, "self");
 
 	vty_out (vty, "%s", VTY_NEWLINE);
       }
@@ -3382,6 +3426,10 @@ rip_clean ()
       for (i = 0; i < ZEBRA_ROUTE_MAX; i++)
 	if (rip->route_map[i].name)
 	  free (rip->route_map[i].name);
+
+      XFREE (MTYPE_ROUTE_TABLE, rip->table);
+      XFREE (MTYPE_ROUTE_TABLE, rip->route);
+      XFREE (MTYPE_ROUTE_TABLE, rip->neighbor);
       
       XFREE (MTYPE_RIP, rip);
       rip = NULL;
